/* event_handler.c
 *
 *  Copyright (C) 2006 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "event_handler.h"

#define BUFSIZE 4096

static struct list_s *search_node (struct list_s **elist, char **event_id,
				   int size);
static int do_cmd_string (char *action, char **substr, char *cstring,
			  int substr_size);
static void do_action (struct event *ecell, char *sensorname);
static void run_external_cmd (struct event *ecell, char *cmd, char *sensorname);
static void free_event_data (struct list_s **head);
static int check_ignorelist (struct action *acell, char **event_id, int size,
			     int *index);
static int register_event (struct list_s **elist, struct rule *rcell, char *alertsrc,
			   char **event_id, int size);
static void update_event_status (struct list_s *ecell);

static pthread_mutex_t event_m = PTHREAD_MUTEX_INITIALIZER;

int
send_event (struct internal *vars, struct rule *rcell, char *alertsrc, 
            char **event_id, int size)
{
  struct list_s *current;
  struct list_s **elist;
  char logstr[BUFSIZE];
  int index;

  elist = &vars->elist;

  /* First check that event_id has no same substrings that ignore list */
  if (check_ignorelist
      ((struct action *) rcell->aptr, event_id, size, &index))
    {
#if 0
      snprintf (logstr, BUFSIZE,
		"[INFO] %s, rule #%d: action ignored based on criteria: %s",
		vars->sensorname, rcell->rulenr, event_id[index]);
      send_alarm ((const char *) logstr);
#endif
      return 1;
    }

  /* Check if event is registered */
  current = search_node (elist, event_id, size);
  if (current == NULL)
    {
      /* Register new event */
      pthread_mutex_lock (&event_m);
      if (register_event (elist, rcell, alertsrc, event_id, size) < 0)
	{
	  snprintf (logstr, BUFSIZE,
		    "CRITICAL: registering new event failed, malloc() %s",
		    strerror (errno));
	  logwriter ((const char *) logstr);
          pthread_mutex_unlock (&event_m);

	  return -1;
	}
      pthread_mutex_unlock (&event_m);
    }
  else
    {
      /* Event exists, update it's status */
      update_event_status (current);
    }

  return 0;
}

static int
register_event (struct list_s **elist, struct rule *rcell, char *alertsrc,
                char **event_id, int size)
{
  struct list_s *new_node;
  struct event *payload;
  int i;

  /* Add new event */
  new_node = slist_push (elist);

  new_node->payload = (void *) malloc (sizeof (struct event));
  if (new_node->payload == NULL)
    return -1;

  /* Initialize event structure */
  payload = (struct event *) new_node->payload;
  payload->t_start = time (NULL);
  payload->acell = rcell->aptr;
  payload->rcell = rcell;
  payload->event_cnt = 1;
  payload->action_idx = 0;
  payload->id_size = size;

  /* Event is indentified by matched substrings;
   * copy substrings to event structure
   */
  payload->event_id = (char **) malloc (size * sizeof (char *));
  if (payload->event_id == NULL)
    return -1;

  for (i = 0; i < size; i++)
    {
      payload->event_id[i] =
	(char *) calloc (strlen (event_id[i]) + 1, sizeof (char));
      if (payload == NULL)
	return -1;
      strncpy (payload->event_id[i], event_id[i], strlen (event_id[i]));
    }

  payload->alertsrc = (char *) calloc (strlen (alertsrc) + 1, sizeof (char));
  if (payload->alertsrc == NULL)
    return -1;
  strncpy (payload->alertsrc, alertsrc, strlen (alertsrc));

  return 0;
}

/* Update counter of existing event */
static void
update_event_status (struct list_s *ecell)
{
  struct event *payload;

  pthread_mutex_lock (&event_m);
  payload = (struct event *) ecell->payload;
  payload->event_cnt++;
  pthread_mutex_unlock (&event_m);
}

/* mainloop for event handler */
void
monitor_events (struct internal *vars)
{
  struct list_s *current;
  struct list_s *previous;
  struct list_s **elist;
  struct event *payload;
  struct do_action *aptr;
  struct timeval timeout;
  time_t tm;
  int i;
  int aindex;
  int del_event;

  elist = &vars->elist;
  
  while (1)
    {
      /* Check if Nuhe has received SIGTERM, SIGINT or
       * SIGHUP signal */
      if (quit_flag)
	{
	  pthread_mutex_lock (&event_m);
	  free_event_data (elist);
	  slist_delete_list (elist);
	  pthread_mutex_unlock (&event_m);
	  break;
	}

      /* Sleep for a while ... */
      timeout.tv_sec = 0;
      timeout.tv_usec = EVENT_SLEEPTIME;
      select (0, NULL, NULL, NULL, &timeout);

      pthread_mutex_lock (&event_m);
      current = *elist;
      previous = NULL;

      /* Loop through event list */
      while (current != NULL)
	{
	  del_event = 0;
	  payload = (struct event *) current->payload;
	  aindex = payload->action_idx;
	  aptr =
	    (struct do_action *) payload->acell->run_action->element[aindex]->
	    payload;

	  tm = time (NULL);

	  /* Only delete old actions that have positive event_max */
	  if (aptr->event_max > 0 && payload->event_cnt <
	      aptr->event_max && (tm - payload->t_start) >= aptr->t_limit)
	    {
	      /* Time elapsed without maximum amount of events */
	      del_event = 1;
	    }

	  /* Run actions if they have reached event_cnt and time limit has exceeded, also
	   * actions with event_cnt 0 will be runned based on their time limit status
	   */
          else if (aptr->event_max > 0 && (payload->event_cnt > 0
                   && payload->event_cnt >= aptr->event_max
                   && (tm - payload->t_start) <= aptr->t_limit)
                   || (aptr->event_max == 0 && (tm - payload->t_start)
                   >= aptr->t_limit))
	     {
	       /* Trigger action for this event */
	       do_action (payload, vars->sensorname);

	       /* Check if we have subsequent actions */
	       if (payload->action_idx < d_array_get_size (payload->acell->run_action) - 1)
 		 {
		   if (((struct do_action *) payload->acell->run_action->
		        element[aindex + 1]->payload)->event_max > 0)
                     {
                       /* We're not interested in events; subsequent action will be
                        * runned after specified time
                        */
	               payload->event_cnt = 0;
                     }

		   payload->t_start = time (NULL);
		   payload->action_idx++;
		 }
	       else
		 {
		   /* No subsequent actions; delete event from list */
		   del_event = 1;
		 }
	     }

          /* Delete event from list */
	  if (del_event)
	    {
	      /* Free event indentification strings */
	      for (i = 0; i < payload->id_size; i++)
		free (payload->event_id[i]);

              /* Free event_id pointer array and alertsrc */
              free (payload->event_id);
              free (payload->alertsrc);

              /* Remove node from event list */
              if (previous != NULL)
                previous->next = current->next;
              else
                *elist = current->next;

	      /* Free payload struct and event node */
              free (current->payload);
	      free (current);

	      /*pthread_mutex_unlock (&event_m);*/
	      break;
	    }
	  previous = current;
	  current = current->next;
	}
      pthread_mutex_unlock (&event_m);
    }
}

/* Wrapper for do_cmd_string and run_external functions */
static void
do_action (struct event *ecell, char *sensorname)
{
  char cstring[BUFSIZE] = { 0 };
  char logstr[BUFSIZE];

  /* If logthreshold is set to EVENT, we log only when event action is
   * triggered
   */

  /* Generate alarms here */
  if (ecell->acell->ltvalue == EVENT)
    {
      char *rstr;
      char alarmstr[LOGBUF];

      if (ecell->action_idx == 0)
        {
          rstr =
             generate_alarm (ecell->rcell, ecell->alertsrc,
                             sensorname, ecell->event_id[0]);
     
          if (rstr != NULL)
            {
              snprintf (alarmstr, LOGBUF,
                        "%s\nAlert repeated %d times", rstr, ecell->event_cnt);
              send_alarm ((const char *) alarmstr);
              free (rstr);
            }
        }
    }

  if (do_cmd_string
      (((struct do_action *) ecell->acell->run_action->
	element[ecell->action_idx]->payload)->action, ecell->event_id,
       cstring, ecell->id_size) == -1)
    {
      snprintf (logstr, BUFSIZE,
		"[ALERT] %s, error: indexing substring out of bounds, can't execute '%s' command for rule #%d",
		sensorname, ((struct do_action *) ecell->acell->run_action->
		 element[ecell->action_idx]->payload)->action,
		ecell->rcell->rulenr);

      send_alarm ((const char *) logstr);
    }
  else
    run_external_cmd (ecell, cstring, sensorname);
}

/* This function copies action string with modifications to cstring
 * ($0-9 characters are substituded with corresponding Pcre substrings)
 */
static int
do_cmd_string (char *action, char **substr, char *cstring, int substr_size)
{
  int i = 0;
  int x = 0;
  int y;
  int sidx;
     
  while (i < strlen (action))
    {
      if (i >= BUFSIZE)
	return 0;
      if (action[i] == '$' && i < strlen (action) - 1)
	{
	  sidx = atoi (&action[++i]);
	  if (sidx > substr_size - 1)
	    return -1;
	  for (y = 0; y < strlen (substr[sidx]); y++)
	    cstring[x++] = substr[sidx][y];
	}
      else
	cstring[x++] = action[i];
      i++;
    }
  return 0;
}

static void
run_external_cmd (struct event *ecell, char *cmd, char *sensorname)
{
  FILE *fp;
  int rval;
  char logstr[BUFSIZE];

  /* Check what kind of action indentifier we have */
  if (ecell->acell->id == CMD)
    {
      /* Run external command specified by rule action */

      /* FIX this to not use popen()*/
      fp = popen (cmd, "r");
      if (fp == NULL)
	{
	  snprintf (logstr, BUFSIZE,
		    "[ALERT] %s, error: can't execute '%s' for rule #%d, %s", 
                    sensorname, cmd,
		    ecell->rcell->rulenr, strerror (errno));

	  send_alarm ((const char *) logstr);
	}
      else
	{
	  rval = pclose (fp);

	  snprintf (logstr, BUFSIZE,
		    "[ALERT] %s, executed command '%s' for rule #%d with return value of %d",
		    sensorname, cmd, ecell->rcell->rulenr, rval);

	  send_alarm ((const char *) logstr);
	}
    }
#if 0
  else if (ecell->acell->id == SHELL)
    {
      /* TODO */
    }
#endif
}

/* This function validates if event should be ignored (based on ignorelist) */
static int
check_ignorelist (struct action *acell, char **event_id, int size, int *index)
{
  int i;
  int x;

  char *ign_data;

  for (i = 0; i < d_array_get_size (acell->ignore); i++)
    {
      for (x = 0; x < size; x++)
	{
          ign_data = (char *) acell->ignore->element[i]->payload;
	  if (strcmp (ign_data, event_id[x]) == 0)
	    {
	      *index = x;
	      return 1;
	    }
	}
    }
  return 0;
}

/* Searches event list, returns pointer to event structure if found
 * or NULL otherwise
 */
static struct list_s *
search_node (struct list_s **elist, char **event_id, int size)
{
  struct list_s *current;
  struct event *payload;
  int i;
  int flag = 0;

  pthread_mutex_lock (&event_m);
  current = *elist;

  while (current != NULL)
    {
      payload = (struct event *) current->payload;
      for (i = 0; i < size; i++)
	{
	  if (payload->acell->usesubstr[i] == 1)
	    {
	      if (strcmp (payload->event_id[i], event_id[i]) != 0)
		{
		  flag = 0;
		  break;
		}
	      else
		flag = 1;
	    }
	}
      if (flag == 1)
	{
	  /* Event found */
	  pthread_mutex_unlock (&event_m);
	  return current;
	}
      current = current->next;
    }
  pthread_mutex_unlock (&event_m);

  return NULL;
}

static void
free_event_data (struct list_s **head)
{
  struct list_s *current;
  struct list_s *next;
  struct event *cell;
  int i;

  current = *head;
  
  while (current != NULL)
    {
      next = current->next;
      cell = (struct event *) current->payload;

      for (i = 0; i < cell->id_size; i++)
	free (cell->event_id[i]);

      free (cell->alertsrc);
      
      current = next;
    }
}
