/* nuhed.c
 *
 * Copyright (C) 2006 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "nuhed.h"

#define ERRSTATMAX -5

/* Logging mode */
enum logmethod
{
  LOG_NODEMGR = 0,
  LOG_FILE,
  LOG_STDOUT
};

/* File status */
enum fstatus
{
  ACTIVE = 0,
  INACTIVE,
  REOPEN,
  DISABLED
};

/* File info structure */
struct finfo
{
  FILE *fp;
  char *filename;
  int errstat;

  enum fstatus status;
};

static void read_configuration (struct internal *vars);
static void reopen_log_file (struct internal *vars, struct finfo *fstrc);
static int open_log_file (struct internal *vars, struct finfo *fstrc);
static void close_log_file (struct finfo *fstrc, enum fstatus status);
static void monitor_log_files (struct internal *vars, int activelogs);
static void process_data (struct internal *vars, char *filename, char *entry);
static struct rule *check_entry (struct list_s *rlist, char *entry,
                                 char **substr, int *rc);
static int parse_filenames (struct d_array *array, void(fill_array)(struct d_array *logs, char *fname),
                            char *buffer);
static void initialize_logs (struct d_array *logs, char *fname);
static void initialize_rules (struct d_array *rules, char *fname);
static void get_timestamp (char *str);
static void daemonize ();
static void exit_cleanly (struct internal *vars);
static void free_rule_data (struct list_s **head);
static void free_action_data (struct list_s **head);
static void sigterm_handler (int signo);

static FILE *logfp = NULL;
static enum logmethod logmode = LOG_FILE;

int
main (int argc, char **argv)
{
  FILE *pidf;
  struct stat buf;
  struct internal vars;
  struct finfo *fstrc;
  int i;
  int ich;
  int trtn;
  int fcnt = 0;

  memset (vars.manageraddr, 0, sizeof (vars.manageraddr));
  memset (vars.sensorname, 0, sizeof (vars.sensorname));
  memset (vars.logfile, 0, sizeof (vars.logfile));
  memset (vars.pidfile, 0, sizeof (vars.pidfile));

  vars.cfile = NULL;
  vars.rlist = NULL;
  vars.alist = NULL;
  vars.elist = NULL;
  vars.rules = NULL;

  vars.run_stat = DAEMONIC;

  opterr = 0;
  while ((ich = getopt (argc, argv, "fvhrsc:")) != -1)
    switch (ich)
      {
        case 'f':
	  vars.run_stat = FOREGROUD;
	  break;
        case 'v':
	  printf ("Nuhe log monitor version %s\n", VERSION);
	  exit (0);
	  break;
        case 'h':
	  printf
	    ("Nuhe - action capable log file monitor %s\nCopyright (C) 2006 Tuomo Makinen\n\nUsage:\n  -c  <file>  Configuration file for Nuhe log monitor\n\n  -f  Foreground mode\n  -r  Redirect logging to stdout\n  -s  <file1> <file2> ...  Log files to scan\n  -v  Show version of Nuhe log monitor\n  -h  Prints this message\n\n", VERSION);
	  exit (0);
	  break;
        case 'c':
          vars.cfile = optarg;
          break;
        case 'r':
          logmode = LOG_STDOUT;
          break;
        case 's':
          vars.run_stat = SCANLOGS;
          logmode = LOG_STDOUT;
          break; 
        case '?':
          fprintf (stderr, "Error: unknown option, -%c\n", optopt);
          exit(1);
          break;
      }

  if (vars.run_stat == DAEMONIC && logmode == LOG_STDOUT)
    {
      fprintf (stderr, "Error: you can only specify stdout logging (-s) with foreground mode (-f)\n");
      exit (1);
    }

  /* Create d_array for finfo structures */
  vars.logs = d_array_create ();

  /* Create d_array for rulefiles */
  vars.rules = d_array_create ();

  /* Read configuration file */
  read_configuration (&vars);

  if (vars.run_stat != SCANLOGS)
    {
     /* Check if the pid file exists; we do not want run
      * multiple instances */
      if ((stat (vars.pidfile, &buf) != -1))
        {
          fprintf (stderr, "Error: PID file exists (%s)\n", vars.pidfile);
          exit (1);
        }
    }

  /* Parse rules */
  parse_rules (&vars.rlist, &vars.alist, vars.rules);

  /* Delete rules d_array */
  d_array_delete (vars.rules);

  if (vars.run_stat == SCANLOGS && optind < argc)
    {
      while (optind < argc)
        initialize_logs (vars.logs, argv[optind++]);
    }

  /* Open monitored log files */
  for (i = 0; i < d_array_get_size (vars.logs); i++)
    {
      fstrc = (struct finfo *) vars.logs->element[i]->payload;

      /* no more files to monitor */
      if (fstrc == NULL)
	break;

      if (open_log_file (&vars, fstrc) == -1)
	{
	  fprintf (stderr, "Skipping file: %s, %s\n",
		   fstrc->filename, strerror (errno));
	  fstrc->status = DISABLED;
	  fstrc->errstat = ERRSTATMAX;
	}
      else
	{
	  fstrc->status = ACTIVE;
	  fstrc->errstat = 0;
	  fcnt++;
	}
    }

  /* Oops... no files to monitor */
  if (!fcnt)
    {
      fprintf (stderr, "Error: zero files to monitor, exiting\n");
      exit (1);
    }

  if (vars.run_stat == SCANLOGS)
    monitor_log_files (&vars, fcnt);
 
  /* DAEMONIC and FOREGROUND specific initializations 
   * ************************************************
   */ 

  /* umask for pid & log files */
  umask (077);

  /* Create PID file */
  pidf = fopen (vars.pidfile, "w");
  if (pidf == NULL)
    {
      fprintf (stderr, "Error: can't create %s, %s\n",
	       vars.pidfile, strerror (errno));
      exit (1);
    }

  /* Open logging file. If we are running standalone mode
   * alerts go to the logfile */
  if (logmode != LOG_STDOUT)
    {
      logfp = fopen (vars.logfile, "a");
      if (logfp == NULL)
        {
          fprintf (stderr,
                   "Error: can't open or create %s, %s\n",
                   vars.logfile, strerror (errno));
          exit (1);
        }
    }

  if (vars.run_stat == DAEMONIC)
    {
      /* Daemonize */
      printf ("Forking to background\n");
      daemonize ();
    }

  /* Set quit flag */
  quit_flag = 0;
 
  /* Initialize signal handlers */
  signal (SIGTERM, sigterm_handler);
  signal (SIGINT, sigterm_handler);
  signal (SIGHUP, sigterm_handler);

  /* Write PID file */
  fprintf (pidf, "%d\n", (int) getpid ());
  fclose (pidf);

  /* Log Nuhe startup */
  if (vars.run_stat == DAEMONIC)
    logwriter ("Nuhe daemon started");
  else
    logwriter ("Nuhe started in foreground mode");

  /* Start event handler thread */
  if ((trtn =
       pthread_create (&vars.event_t, NULL, (void *) monitor_events, &vars)))
    {
      fprintf (stderr, "Error: pthread_create, %s\n", strerror (trtn));
      exit (1);
    }

  /* Monitor log files forever */
  monitor_log_files (&vars, fcnt);

  return 0;
}

/* daemonizer */
static void
daemonize ()
{
  pid_t pid;

  pid = fork ();
  if (pid < 0)
    {
      perror ("fork() failed");
      exit (errno);
    }
  else if (pid != 0)
    exit (0);
  else
    {
      setsid ();
      chdir ("/");

      /* Close stdin, stdout and stderr */
      close (STDIN_FILENO);
      close (STDOUT_FILENO);
      close (STDERR_FILENO);
    }
}

/* This function parse multiple filenames from configuration
 * file and calls fill_array function pointer to add it to d_array 
 */
static int
parse_filenames (struct d_array *array, void (fill_array) (struct d_array *logs, char *fname),
                 char *buffer)
{
  char *cptr;
  char *lfile;
  char fname[4096] = {
    0
  };

  cptr = &buffer[0];
  lfile = &fname[0];

  while (isspace ((int) *cptr))
    cptr++;
  while (isgraph ((int) *cptr) || *cptr == '/')
    *lfile++ = *cptr++;
  while (isspace ((int) *cptr))
    *cptr++;

  (*fill_array) (array, fname);

  if (*cptr == '\\')
    return 0;
  else
    return 1;
}

/* Initializes logs d_array with finfo structs */
static void
initialize_logs (struct d_array *logs, char *fname)
{
  int index;

  index = d_array_add_element (logs);

  logs->element[index]->payload = (void *) malloc (sizeof (struct finfo));
  if (logs->element[index]->payload == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  ((struct finfo *) logs->element[index]->payload)->fp = NULL;
  ((struct finfo *) logs->element[index]->payload)->status = INACTIVE;
  ((struct finfo *) logs->element[index]->payload)->filename =
     (void *) calloc (strlen (fname) + 1, sizeof (char));
  if (((struct finfo *) logs->element[index]->payload)->filename == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  strcpy (((struct finfo *) logs->element[index]->payload)->filename, fname);
}

/* Initializes rules d_array with rule filenames */
static void
initialize_rules (struct d_array *rules, char *fname)
{
  int index;
 
  index = d_array_add_element (rules);
  
  rules->element[index]->payload = (void *) calloc (strlen (fname) + 1, sizeof (char));
  if (rules->element[index]->payload == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  strcpy ((char *) rules->element[index]->payload, fname);
}

void
read_configuration (struct internal *vars)
{
  FILE *fp;
  char buffer[1024];
  char *cptr;
  int rlogs = 1;
  int smode = 1;
  int i = 0;

  struct d_array *dptr = NULL;
  void (*fill_array) (struct d_array *logs, char *fname) = NULL;
  char *confiles[] = {
    "../etc/nuhed/nuhed.conf",
    "/usr/local/etc/nuhed.conf",
    "/usr/local/etc/nuhed/nuhed.conf",
    "/etc/nuhed/nuhed.conf", "/etc/nuhed.conf",
    "./nuhed.conf"
  };

  if (vars->cfile == NULL)
    {
      for (i = 0; i < 6; i++)
        {
          fp = fopen (confiles[i], "r");
          if (fp != NULL)
	    break;
        }
    }
  else
    fp = fopen (vars->cfile, "r");

  if (fp == NULL)
    {
      fprintf (stderr,
	       "Error: can't open configuration file %s, %s\n",
	       vars->cfile == NULL ? confiles[i] : vars->cfile, strerror (errno));
      exit (1);
    }

  printf ("Using %s configuration file\n", vars->cfile == NULL ? confiles[i] : vars->cfile);

  i = 0;
  while (fgets (buffer, sizeof (buffer), fp) != NULL)
    {
      if (!strncmp (buffer, "#", 1))
	continue;
      else if (*buffer == '\0')
	continue;
      else if (rlogs == 0)
        {
          if (vars->run_stat == SCANLOGS && dptr == vars->logs)
            break;
          else
            rlogs = parse_filenames (dptr, fill_array, buffer);
        }
      else if (!strncmp (buffer, "Rules", 5))
        {
          fill_array = &initialize_rules;
          dptr = vars->rules;
          rlogs = -1;
        }
      else if (vars->run_stat != SCANLOGS && !strncmp (buffer, "Logs", 4))
	{
          fill_array = &initialize_logs;
          dptr = vars->logs;
          rlogs = -1;
	}
      else if (!strncmp (buffer, "Standalone", 10)
               && logmode != LOG_STDOUT)
        {
	  sscanf (buffer, "Standalone %d", &smode);
          if (!smode)
            logmode = LOG_NODEMGR;
        }
      else if (!strncmp (buffer, "Pidfile", 7))
	sscanf (buffer, "Pidfile %255s", vars->pidfile);
      else if (!strncmp (buffer, "Logfile", 7))
	sscanf (buffer, "Logfile %255s", vars->logfile);
      else if (!strncmp (buffer, "NodemanagerAddr", 14))
	sscanf (buffer, "NodemanagerAddr %255s", vars->manageraddr);
      else if (!strncmp (buffer, "NodemanagerPort", 14))
	sscanf (buffer, "NodemanagerPort %u", &vars->managerport);
      else if (!strncmp (buffer, "NodeName", 10))
	sscanf (buffer, "NodeName %255s", vars->sensorname);

      /* read first string from multiple lines (Rules and Logs) */
      if (rlogs == -1)
        {
          cptr = &buffer[0];
          while (!isspace ((int) *cptr))
            cptr++;
          cptr++;
          rlogs = parse_filenames (dptr, fill_array, cptr);
        }
    }
  fclose (fp);

  if (logmode == LOG_NODEMGR)
    {
      fprintf (stderr,
	       "Error: only standalone configuration is supported at the moment.\n");
      exit (1);
    }
  if (vars->rules == NULL)
    {
      fprintf (stderr,
	       "Error reading configuration file: no rule files specified with Rule\n");
      exit (1);
    }
  if (*vars->pidfile == '\0')
    {
      strncpy (vars->pidfile, "/var/run/nuhed.pid", STRBUF);
    }
  if (*vars->logfile == '\0')
    {
      fprintf (stderr,
	       "Error reading configuration file: You must specify logfile\n");
      exit (1);
    }
  if (logmode == LOG_NODEMGR && *vars->manageraddr == '\0')
    {
      fprintf (stderr,
	       "Error reading configuration file: no nodemanager address specified with NodemanagerAddr\n");
      exit (1);
    }
  if (vars->run_stat != SCANLOGS && rlogs != 1)
    {
      fprintf (stderr,
	       "Error reading configuration file: no monitored files specified with Logs\n");
      exit (1);
    }
  if (*vars->sensorname == '\0')
    {
      /* Use system name for sensorname */
      struct utsname buf;
      if (uname (&buf) == -1)
	{
	  fprintf (stderr,
		   "Error: no SensorName specified and uname() failed, %s\n",
		   strerror (errno));
	  exit (1);
	}
      strncpy (vars->sensorname, buf.nodename, STRBUF);
    }
}

/* Handle log file opening */
static int
open_log_file (struct internal *vars, struct finfo *fstrc)
{
  fstrc->fp = fopen (fstrc->filename, "r");

  if (fstrc->fp == NULL)
    return -1;
  else
    {
      /* seek to end of the file, if file is
       * opened during daemon startup.
       */
      if (vars->run_stat != SCANLOGS && fstrc->status == INACTIVE)
	{
	  if (fseek (fstrc->fp, 0, SEEK_END) < 0)
	    {
	      fclose (fstrc->fp);
	      fstrc->fp = NULL;
	      return -1;
	    }
	}
      return 1;
    }
}

static void
close_log_file (struct finfo *fstrc, enum fstatus status)
{
  if (fstrc->fp != NULL)
    {
      fclose (fstrc->fp);
      fstrc->fp = NULL;
    }
  fstrc->status = status;
}

/* This function tries to reopen log file and update it's status */
static void
reopen_log_file (struct internal *vars, struct finfo *fstrc)
{
  char alarmstr[LOGBUF];
  int rval;

  rval = open_log_file (vars, fstrc);

  if (rval == -1)
    {
      snprintf (alarmstr, LOGBUF,
		"[CRIT] %s: Can't reopen log file %s, %s",
		vars->sensorname, fstrc->filename, strerror (errno));
      send_alarm ((const char *) alarmstr);
      fstrc->errstat--;
    }
  else
    {
      snprintf (alarmstr, LOGBUF,
		"[INFO] %s: File %s reopened",
		vars->sensorname, fstrc->filename);
      send_alarm ((const char *) alarmstr);
      fstrc->status = ACTIVE;
      fstrc->errstat = 0;
    }
}

/* Main loop for Nuhe: monitor logfiles, check files against rules
 * and generate alerts */
static void
monitor_log_files (struct internal *vars, int activelogs)
{
  char buffer[LOGBUF];
  char alarmstr[LOGBUF];
  struct stat sbuf, fbuf;
  struct timeval timeout;
  struct finfo *fstrc;
  long pos;
  int i;

  while (1)
    {
      /* No more files to monitor ... */
      if (!activelogs)
	{
	  snprintf (alarmstr, LOGBUF,
		    "[CRIT] %s: no files to monitor, exiting",
		    vars->sensorname);
	  send_alarm ((const char *) alarmstr);
	  quit_flag = 1;
	}

      for (i = 0; i < d_array_get_size (vars->logs); i++)
	{
	  fstrc = (struct finfo *) vars->logs->element[i]->payload;

          /* Workaround for -s command line option */
          if (vars->run_stat == SCANLOGS)
            {
              if (fstrc->status != DISABLED)
                {
                  printf ("*** Analyzing file %s ***\n\n", fstrc->filename);
                  while (1)
                    {
                      if (fgets (buffer, sizeof (buffer), fstrc->fp) == NULL)
                        break;
                      else
                        process_data (vars, fstrc->filename, buffer);
                    }
                }
              if (i == d_array_get_size (vars->logs) - 1)
                {
                  exit_cleanly (vars);
                  exit (0);
                }
              continue;  
            }

          timeout.tv_sec = 0;
          timeout.tv_usec = FMON_SLEEPTIME;

	  /* If quit_flag is set to 1 we have received SIGTERM, SIGINT or SIGHUP
           * and should exit cleanly */
	  if (quit_flag)
	    exit_cleanly (vars);

	  /* Break from loop if uninitialized element is reached */
	  if (fstrc == NULL)
	    break;

	  /* Just jump away from disabled or inactive files */
	  if (fstrc->status == DISABLED || fstrc->status == INACTIVE)
	    continue;
	  else if (fstrc->status == REOPEN)
	    {
	      /* File is marked to be reopened... this indicates that file
	       * has possibly rotated; try to reopen it
	       */
	      if (fstrc->errstat == ERRSTATMAX)
		{
		  /* file will be disabled; too many errors */
		  snprintf (alarmstr, LOGBUF,
			    "[CRIT] %s: too many errors with %s file, disabling it from monitoring",
			    vars->sensorname, fstrc->filename);
		  send_alarm ((const char *) alarmstr);
		  close_log_file (fstrc, DISABLED);
		  activelogs--;
		}
	      else
		{
		  /* sleep for a while, before reopening file */
                  timeout.tv_sec = 1;
                  timeout.tv_usec = 0;
		  select (0, NULL, NULL, NULL, &timeout);
		  reopen_log_file (vars, fstrc);
		}

	      continue;
	    }
	  
          /* Do some validation checks for monitored files */
	  fstat (fileno (fstrc->fp), &fbuf);
	  if (fbuf.st_nlink == 0)
	    {
	      snprintf (alarmstr, LOGBUF,
			"[ALERT] %s: Someone removed %s file (probably log is rotated)\
, hard link count reached zero, trying to reopen file...",
			vars->sensorname, fstrc->filename);
	      send_alarm ((const char *) alarmstr);
	      close_log_file (fstrc, REOPEN);
	      continue;
	    }

	  if (fbuf.st_nlink > 0 && stat (fstrc->filename, &sbuf) != -1)
	    {
	      if (sbuf.st_ino != fbuf.st_ino)
		{
		  snprintf (alarmstr, LOGBUF,
			    "[ALERT] %s: Filename %s inode has changed.\
 File exists with different name and someone has created another file with %s name, trying to reopen file...",
			    vars->sensorname, fstrc->filename, fstrc->filename);
		  send_alarm ((const char *) alarmstr);
		  close_log_file (fstrc, REOPEN);
		  continue;
		}
	    }
	  else
	    {
	      snprintf (alarmstr, LOGBUF,
			"[ALERT] %s: File %s has positive hard link count, \
but it's filename does not exists. this probably means that someone has renamed file, trying to reopen file...",
			vars->sensorname, fstrc->filename);
	      send_alarm ((const char *) alarmstr);
	      close_log_file (fstrc, REOPEN);
	      continue;
	    }

          /* Read new line from log file and process data */
	  pos = ftell (fstrc->fp);
	  if (fgets (buffer, sizeof (buffer), fstrc->fp) == NULL)
	    {
	      fseek (fstrc->fp, pos, SEEK_SET);
	      clearerr (fstrc->fp);
	      select (0, NULL, NULL, NULL, &timeout);
	    }
	  else
            process_data (vars, fstrc->filename, buffer);
        }
    }
}

static void
process_data (struct internal *vars, char *filename, char *entry)
{
  struct rule *matched = NULL;
  char *substr[SUBSTRMAX];
  char *alarmstr;
  int rc;
  int rc_cnt;

  matched = check_entry (vars->rlist, entry, substr, &rc);

  if (matched != NULL)
    {
      /* PCRE matched */
      if (matched->aptr->loglevel != NONE &&
          (matched->aptr->ltvalue == ALWAYS ||
           vars->run_stat == SCANLOGS))
        {
          /* Generate alarm */
          alarmstr =
             generate_alarm (matched, filename,
                             vars->sensorname, entry);

          if (alarmstr != NULL)
            {
              send_alarm ((const char *) alarmstr);
              free (alarmstr);
            }
        }

      /* process action if specified */
      if (vars->run_stat != SCANLOGS && matched->aptr->id != EMPTY)
         {
           /* Actions are handled by event handler */
           send_event (vars, matched, filename, substr, rc);
         }

      /* Free substrings */
      for (rc_cnt = 0; rc_cnt < rc; rc_cnt++)
        if (substr[rc_cnt] != NULL)
           free (substr[rc_cnt]);
    }
}
 
static struct rule *
check_entry (struct list_s *rlist, char *entry, char **substr, int *rc)
{
  struct list_s *current;
  struct rule *payload;
  int ovector[OVECCOUNT];
  int i;

  current = rlist;

  while (current != NULL)
    {
      payload = (struct rule *) current->payload;
      *rc = pcre_exec (payload->re,
		       NULL,
		       entry, (int) strlen (entry), 0, 0, ovector, OVECCOUNT);

      /* Collect substrings */
      if (*rc > 0)
	{
	  for (i = 0; i < *rc; i++)
	    {
	      char *substring_start = entry + ovector[2 * i];
	      int substring_length = ovector[2 * i + 1] - ovector[2 * i];
	      substr[i] =
		(char *) calloc ((substring_length + 1), sizeof (char));
	      strncpy (substr[i], substring_start, substring_length);
	    }
	}

      if (*rc > 0)
	{
	  return payload;
	}
      current = current->next;
    }
  return NULL;
}

static void
sigterm_handler (int signo)
{
  quit_flag = 1;

  if (signo == SIGTERM)
    logwriter ("Received the SIGTERM signal - exiting");
  else if (signo == SIGINT)
    logwriter ("Received the SIGINT signal - exiting");
  else
    logwriter ("Received the SIGHUP signal - exiting");
}

static void
exit_cleanly (struct internal *vars)
{
  struct finfo *fstrc;
  struct d_array *logs;
  int darr_sz;
  int i;

  logs = vars->logs;

  /* Wait event handler thread to exit */
  if (vars->run_stat != SCANLOGS)
    pthread_join (vars->event_t, NULL);

  /* Free rule and action structs */
  free_rule_data (&vars->rlist);
  free_action_data (&vars->alist);

  /* Free rule and action lists */
  slist_delete_list (&vars->rlist);
  slist_delete_list (&vars->alist);

  darr_sz = d_array_get_size (logs);
  /* Close active fd's and free memory that is allocated
   * for finfo structures
   */
  for (i = 0; i < darr_sz; i++)
    {
      fstrc = (struct finfo *) logs->element[i]->payload;
      close_log_file (fstrc, DISABLED);
      free (fstrc->filename);
    }
  
  /* Delete logs d_array */
  d_array_delete (logs);

  if (vars->run_stat != SCANLOGS)
    {
      if (unlink (vars->pidfile) != 0)
        logwriter ("Error: can't unlink PID file");
    }
 
  if (logmode != LOG_STDOUT) 
    fclose (logfp);

  exit (0);
}

static void
free_rule_data (struct list_s **head)
{
  struct list_s *current;
  struct list_s *next;
  struct rule *cell;

  current = *head;

  while (current != NULL)
    {
      next = current->next;
      cell = (struct rule *) current->payload;
      if (cell->actname != NULL)
        free (cell->actname);

      free (cell->desc);
      free (cell->re);
      current = next;
    }
}

static void
free_action_data (struct list_s **head)
{
  struct list_s *current;
  struct list_s *next;
  struct action *cell;
  unsigned int darr_sz;
  int i;

  current = *head;

  while (current != NULL)
    {
      next = current->next;
      cell = (struct action *) current->payload;

      /* Delete ignore d_array */
      if (cell->ignore != NULL)
        d_array_delete (cell->ignore);

      if (cell->run_action != NULL)
        {
          /* Free action from do_action struct */
          darr_sz = d_array_get_size (cell->run_action);
          for (i = 0; i < darr_sz; i++)
            free (((struct do_action *)cell->run_action->element[i]->payload)->action);

          /* Free d_array */
          d_array_delete (cell->run_action);
        }
      
      free (cell->name);
      current = next;
    }
}

/* Supports only 24h clock */
static void
get_timestamp (char *str)
{
  struct tm *tmptr;
  time_t tm;

  tm = time (NULL);
  tmptr = (struct tm *) localtime (&tm);

  strftime (str, 1024, "%b %d %H:%M:%S %Y", tmptr);
}

char *
generate_alarm (struct rule *matched, char *src, char *sensorname, char *data)
{
  char loglevel[6] = {
    0
  };
  char *alarmstr;

  if (matched->aptr->loglevel == CRIT)
    strncpy (loglevel, "CRIT", 4);
  else if (matched->aptr->loglevel == ALERT)
    strncpy (loglevel, "ALERT", 5);
  else if (matched->aptr->loglevel == INFO)
    strncpy (loglevel, "INFO", 4);

  if (matched->aptr->loglevel != NONE)
    {
      alarmstr = (char *) calloc (LOGBUF, sizeof (char));
      if (alarmstr == NULL)
        logwriter ("CRITICAL: can't allocate memory for alarm string");
      else
        {
          if (data[strlen (data) - 1] == '\n')
            data[strlen (data) - 1] = '\0';

          snprintf (alarmstr, LOGBUF,
                    "[%s] %s, rule #%d: %s, received from %s. Matched content: %s",
                    loglevel, sensorname, matched->rulenr,
                    matched->desc, src, data);
        }
    }
  else
    alarmstr = NULL;

  return alarmstr;
}

void
send_alarm (const char *alarmstr)
{
  if (logmode != LOG_NODEMGR)
    {
      /* Just log to normal log file */
      logwriter (alarmstr);
    }
  else
    {
      /* Send alarms to nodemanager
       * To be implemented */
    }
}

void
logwriter (const char *log_entry)
{
  char str[1024];

  get_timestamp (str);
  
  fprintf (logmode == LOG_FILE ? logfp : stdout,
           "[%s][%d] %s\n", str, (int) getpid (), log_entry);
  fflush (logfp);
}
