/* ruleparser.c
 *
 *  Copyright (C) 2006 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "ruleparser.h"

#define BUFSIZE 1024

/* ruleparser states */
enum states
{
  GLOBAL_SPACE = 0,
  RULES_SPACE,
  ACTION_SPACE,
  AFUNC_SPACE,
  GROUP_SPACE
};

static void arrange_nodes (struct list_s **rlist, struct list_s **alist);
static struct action *allocate_action_node (struct list_s *node, char *bptr);
static void __parse_rules (struct list_s **rlist, struct list_s **alist, const char *filename,
                          int *rulenumber);
static int parse_rules_space (struct list_s **rlist, enum states pflag, char *bptr,
              		      int rulenumber, int line, int groupread);
static int parse_afunc_space (struct action *new_node, char *bptr,
			      int funcread, int line);
static void copy_string (char *bptr, char *cptr);
static int parse_ignorelist (struct d_array *ignore, char *buffer);
static int do_substr_table (struct action *new_node, char *bptr);
static void initialize_action (struct action *acell);

static int read_ignorelist = 1;
static int action_allocated = 0;
static int old_index = -1;

/* Wrapper for __parse_rules */
void
parse_rules (struct list_s **rlist, struct list_s **alist, struct d_array *rules)
{
  int rulenumber = 0,
      i;

  for (i = 0; i < d_array_get_size (rules); i++)
    {
      __parse_rules (rlist, alist,
                   (const char *) rules->element[i]->payload, &rulenumber);
    }

  /* Adjust rule nodes to point correct action nodes */
  arrange_nodes (rlist, alist);
}

void
__parse_rules (struct list_s **rlist, struct list_s **alist,
              const char *filename, int *rulenumber)
{
  FILE *fp;
  int line = 0;
  int val;
  int rulesread = -1;
  int actionread = -1;
  int funcread = -1;
  int groupread = -1;
  char buffer[BUFSIZE];
  char *bptr;
  enum states pflag = GLOBAL_SPACE;
  struct list_s *new_node = NULL;
  struct action *payload;

  fp = fopen (filename, "r");
  if (fp == NULL)
    {
      fprintf (stderr,
	       "Error: can't open specified rule file (%s).\n",
	       filename);
      exit (1);
    }

  printf ("Using %s rule file\n", filename);

  while (fgets (buffer, sizeof (buffer), fp) != NULL)
    {
      line++;

      bptr = &buffer[0];

      while (isspace ((int) *bptr))
        bptr++;

      /* Just jump away from comments and empy lines */
      if ((*bptr == '#') || (*bptr == '\0'))
	continue;

      if ((pflag == RULES_SPACE) || (pflag == ACTION_SPACE))
	{
	  if ((actionread == 0 || rulesread == 0) && *bptr != '{')
	    {
	      fprintf (stderr, "Error: parse error at line %d\n", line);
	      exit (1);
	    }
	  else
	    pflag == RULES_SPACE ? rulesread++ : actionread++;

	  if (*bptr == '{')
	    continue;
	  else if (*bptr == '}')
	    {
	      pflag = GLOBAL_SPACE;
	      continue;
	    }
	}

      /* Scan for "rules" indentifier */
      if (!strncmp (bptr, "rules", 5) && pflag == GLOBAL_SPACE)
	{
	  if (rulesread >= 0)
	    {
	      fprintf (stderr, "Error: redefinition of rules at line %d\n",
		       line);
	      exit (1);
	    }
	  rulesread++;
	  pflag = RULES_SPACE;
	  continue;
	}
      /* Scan for "action" indentifier */
      else if (!strncmp (bptr, "action", 6) && pflag == GLOBAL_SPACE)
	{
	  if (actionread >= 0)
	    {
	      fprintf (stderr, "Error: redefinition of action at line %d\n",
		       line);
	      exit (1);
	    }
	  actionread++;
	  pflag = ACTION_SPACE;
	  continue;
	}
      else if (pflag == GLOBAL_SPACE)
	{
	  fprintf (stderr, "Error: parse error at line %d\n", line);
	  exit (1);
	}

      /* Start parsing ... */
      if (pflag == RULES_SPACE)
	{
          rulesread++;
          val = parse_rules_space (rlist, pflag, bptr, *rulenumber, line, groupread);
	  if (val == 0)
	    (*rulenumber)++;
	  else if (val == 1)
            {
              /* Inside group; rules need to be grouped */
              pflag = GROUP_SPACE;
            }
	}
      else if (pflag == GROUP_SPACE)
        {
          groupread++;
          val = parse_rules_space (rlist, pflag, bptr, *rulenumber, line, groupread);
          if (val == -1)
            {
              groupread = -1;
              pflag = RULES_SPACE;
            }
          else if (val == 0)
            (*rulenumber)++;
        }
      else if (pflag == ACTION_SPACE)
	{
	  new_node = slist_append_node (alist);
	  payload = allocate_action_node (new_node, bptr);
	  actionread++;
	  pflag = AFUNC_SPACE;
	}
      else if (pflag == AFUNC_SPACE)
	{
	  funcread++;
	  val =
	    parse_afunc_space ((struct action *) new_node->payload, bptr,
			       funcread, line);
	  if (val == -1)
	    {
	      funcread = -1;
	      pflag = ACTION_SPACE;
	      action_allocated = 0;
	    }
	}
    }

  if (pflag != GLOBAL_SPACE)
    {
      fprintf (stderr, "Error: parse error at line %d\n", line);
      exit (1);
    }

  fclose (fp);
}

static int
parse_rules_space (struct list_s **rlist, enum states pflag, char *bptr, int rulenumber,
		   int line, int groupread)
{
  struct list_s *current = NULL;
  struct rule *payload;
  int i = 0;
  int len;
  int chk = -1;
  int erroffset;
  int offset[4] = { 0 };
  char *rulestr = NULL;
  char *cptr;
  char **cptrptr;
  const char *error;

  /* Adjust pointers to correct strings */
  cptr = &bptr[0];
  while (!isspace ((int) *bptr))
    bptr++;
  *bptr = '\0';
 
  /* We must handle group of rules */ 
  if (strncmp (cptr, "group", 5) == 0)
    return 1; 

  /* Do validation and check status of group block */
  if (pflag == GROUP_SPACE)
    {
      if (*cptr != '{' && groupread == 0)
        {
          fprintf (stderr, "Error: parse error at line %d\n", line);
          exit (1);
        }
      else if (*cptr == '}' && groupread == 1)
        {
          fprintf (stderr, "Error: group ends empty at line %d\n", line);
          exit (1);
        }
      else if (*cptr == '{')
        return 1;
      else if (*cptr == '}')
        return -1;
    }

  /* Add node and allocate space for rule struct */ 
  current = slist_append_node (rlist);
  current->payload = (void *) malloc (sizeof (struct rule));

  /* Mark rule number */
  payload = (struct rule *) current->payload;
  payload->rulenr = rulenumber;

  /* actname indicates action name for rule */
  payload->actname = (char *) calloc (strlen (cptr) + 1, sizeof (char));
  if (payload->actname == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  strncpy (payload->actname, cptr, strlen (cptr));

#if 0
  if (pflag != GROUP_SPACE)
    payload->group = NULL;
  else if (pflag == GROUP_SPACE && groupread == 0)
    {
      payload->group = (struct rulegroup *) malloc (sizeof (struct rulegroup));
    }
#endif

  /* Trim whitespaces and tabs away */
  bptr++;
  while (isspace ((int) *bptr))
    bptr++;

  /* Build string indexing table */
  len = strlen (bptr);
  while (len >= 0)
    {
      if (bptr[len] == '\"')
	{
	  chk++;
	  if ((bptr[len - 1] > 0 && chk == 3))
	    chk--;
	  else
	    offset[i++] = len;
	}
      else if ((chk == 1 && !isspace ((int) bptr[len])) ||
               (len == 0 && bptr[len] != '\"'))
	{
	  fprintf (stderr, "Error: parse error at line %d\n", line);
	  exit (1);
	}
      len--;
    }

  /* Allocate space for rule description and regexp and copy
   * corresponding strings to them */
  i = 0;
  cptrptr = &payload->desc;
  while (1)
    {
      cptr = &bptr[offset[i + 1]];
      len = (offset[i]) - offset[i + 1];
      *cptrptr = (char *) calloc (len, sizeof (char));
      if (*cptrptr == NULL)
        {
          perror ("calloc() failed");
          exit (1);
        }

      strncpy (*cptrptr, ++cptr, len - 1);
      if (i)
        break;
      cptrptr = &rulestr;
      i += 2;
    }
  
  /* Do validation for rule node members */
  if (*rulestr == '\0')
    {
      fprintf (stderr,
	       "Error: no search string specified for %s at line %d\n",
	       payload->actname, line);
      exit (1);
    }
  if (*payload->desc == '\0')
    {
      fprintf (stderr,
	       "Error: no description for rule %s specified at line %d\n",
	       payload->actname, line);
      exit (1);
    }

  /* Okay, compile regex from rulestr and check that compilation is succesful */
  payload->re = pcre_compile (rulestr, 0, &error, &erroffset, NULL);
  free (rulestr);

  if (payload->re == NULL)
    {
      fprintf (stderr,
	       "Error: failed to compile regular expression for %s rule\n",
	       payload->actname);
      exit (1);
    }
  
  return 0;
}

static int
parse_afunc_space (struct action *new_node, char *bptr, int funcread,
		   int line)
{
  char *firststr;
  int index = -1;
  struct do_action *ahandler;
  
  if (*bptr != '{' && funcread == 0)
    {
      fprintf (stderr, "Error: parse error at line %d\n", line);
      exit (1);
    }
  else if (*bptr == '}' && funcread == 1)
    {
      fprintf (stderr, "Error: action ends empty at line %d\n", line);
      exit (1);
    }
  else if (*bptr == '{')
    return 0;
  else if (*bptr == '}')
    {
      /* Do some validation for run_actions */
      if (new_node->id != EMPTY)
	{
	  /* Loop through run_action array */
	  struct do_action *ahandler;
	  int i;
	  
          for (i = 0; i < d_array_get_size (new_node->run_action); i++)
	    {
	      ahandler = (struct do_action *) new_node->run_action->element[i]->payload;
	      if (ahandler->event_max == -1)

		{
		  fprintf (stderr,
			   "Error: at line %d, eventcount must be specified for action\n",
			   line);
		  exit (1);
		}
              else if (ahandler->t_limit == -1)
                {
                  fprintf (stderr,
                           "Error: at line %d, timelimit must be specified for action\n",
                           line);
                  exit (1);

                }
              else if (ahandler->action == NULL)
                {
                  fprintf (stderr,
                           "Error: at line %d, command must be specified for action\n",
                           line);
                  exit (1);
                }
	      else if (!read_ignorelist)
		{
		  fprintf (stderr,
			   "Error: at line %d, unterminated ignorelist\n",
			   line);
		  exit (1);
		}
	    }
	}
      return -1;
    }
  else
    {
      /* Just read subsequent string from ignore list */
      if (!read_ignorelist)
        {
          read_ignorelist = parse_ignorelist (new_node->ignore, bptr);
          return 0;
        }

      firststr = &bptr[0];
      while (!isspace ((int) *bptr))
        bptr++;
      *bptr = '\0';
      bptr++;

      if (strncmp (firststr, "command", 7) == 0)
	{
	  new_node->id = CMD;

	  sscanf (firststr, "command[%d]", &index);
	  if (index > old_index)
	    action_allocated = 0;

	  if (index < 0 || (action_allocated == 1
	      && ((struct do_action *) new_node->run_action->element[index]->payload)->action
              != NULL) || (action_allocated == 0 && (index > old_index + 1
	      || index < old_index)))
	    {
	      fprintf (stderr,
		       "Error: at line %d, error indexing command\n", line);
	      exit (1);
	    }

	  if ((new_node->run_action != NULL && d_array_get_size (new_node->run_action) == index)
              || new_node->run_action == NULL)
            initialize_action (new_node);

	  ahandler = (struct do_action *) new_node->run_action->element[index]->payload;

	  /* Allocate space for action string and trim it */
	  ahandler->action =
	    (char *) calloc (strlen (bptr) + 1, sizeof (char));
	  if (ahandler->action == NULL)
	    {
	      perror ("calloc() failed");
	      exit (1);
	    }

          /* Copy string to do_action struct action */
          copy_string (bptr, ahandler->action);

	  if (*ahandler->action == '\0')
	    {
	      fprintf (stderr,
		       "Error: no value for keyword at line %d\n", line);
	      exit (1);
	    }
	}
      else if (strncmp (firststr, "loglevel", 8) == 0)
	{
	  /* Parse logging level */
	  if (strncmp (bptr, "crit", 4) == 0)
	    new_node->loglevel = CRIT;
	  else if (strncmp (bptr, "alert", 5) == 0)
	    new_node->loglevel = ALERT;
	  else if (strncmp (bptr, "info", 4) == 0)
	    new_node->loglevel = INFO;
	  else if (strncmp (bptr, "none", 4) == 0)
	    new_node->loglevel = NONE;
	  else
	    {
	      fprintf (stderr,
		       "Error: at line %d, loglevel value should be crit, alert, info or none\n",
		       line);
	      exit (1);
	    }
	}
      else if (strncmp (firststr, "logthreshold", 12) == 0)
        {
          /* Parse logging treshold */
          if (strncmp (bptr, "event", 5) == 0)
            new_node->ltvalue = EVENT;
          else if (strncmp (bptr, "always", 6) == 0)
            new_node->ltvalue = ALWAYS;
          else
            {
              fprintf (stderr,
                       "Error: at line %d, logthreshold value should be event or always\n",
                       line);
              exit (1);
            }
        }
      else if (strncmp (firststr, "eventcount", 9) == 0)
	{
	  sscanf (firststr, "eventcount[%d]", &index);
	  if (index > old_index)
	    action_allocated = 0;

	  if (index < 0 || (action_allocated == 1 &&
	      ((struct do_action*) new_node->run_action->element[index]->payload)->event_max
              != -1) || (action_allocated == 0 && (index > old_index + 1
              || index < old_index)))
	    {
	      fprintf (stderr,
		       "Error: at line %d, error indexing eventcount\n",
		       line);
	      exit (1);
	    }

          if ((new_node->run_action != NULL && d_array_get_size (new_node->run_action) == index)
              || new_node->run_action == NULL)
            initialize_action (new_node);

	  ahandler = (struct do_action *) new_node->run_action->element[index]->payload;

	  sscanf (bptr, "%d", &ahandler->event_max);
	  if (ahandler->event_max < 0)
	    {
	      fprintf (stderr,
		       "Error: at line %d, eventcount value is less than zero\n",
		       line);
	      exit (1);
	    }
	}
      else if (strncmp (firststr, "timelimit", 9) == 0)
	{
	  sscanf (firststr, "timelimit[%d]", &index);
	  if (index > old_index)
	    action_allocated = 0;
	  
          if (index < 0 || (action_allocated == 1 &&
	      ((struct do_action *) new_node->run_action->element[index]->payload)->t_limit
              != -1) || (action_allocated == 0 && (index > old_index + 1
	      || index < old_index)))
	    {
	      fprintf (stderr,
		       "Error: at line %d, error indexing timelimit\n", line);
	      exit (1);
	    }

          if ((new_node->run_action != NULL && d_array_get_size (new_node->run_action) == index)
              || new_node->run_action == NULL)
            initialize_action (new_node);

          ahandler = (struct do_action *) new_node->run_action->element[index]->payload;
	  sscanf (bptr, "%ld", &ahandler->t_limit);
	 
          if (ahandler->t_limit < 0)
	    {
	      fprintf (stderr,
		       "Error: at line %d, timelimit value must be zero or greater\n",
		       line);
	      exit (1);
	    }
	}
      else if (strncmp (firststr, "ignorelist", 10) == 0 || !read_ignorelist)
	{
	  read_ignorelist = parse_ignorelist (new_node->ignore, bptr);
	}
      else if (strncmp (firststr, "usesubstr", 9) == 0 || !read_ignorelist)
	{
	  if (do_substr_table (new_node, bptr) < 0)
	    {
	      fprintf (stderr, "Error: parse error at line %d\n", line);
	      exit (1);
	    }
	}
      else
	{
	  fprintf (stderr,
		   "Error: wrong value for keyword at line %d\n", line);
	  exit (1);
	}

      old_index = index;
    }
  return 0;
}

static void
copy_string (char *bptr, char *cptr)
{
  int len;
  int i = 0;
  
  len = strlen (bptr);

  while (*bptr != '\0')
    {
      /* Don't copy newlines or starting/ending "'s */
      if ((*bptr == '\n') || 
           (*bptr == '\"' &&
           (i == 0 || i == len - 2)))
	*bptr++;
      else
	*cptr++ = *bptr++;
      i++;
    }
}

static struct action *
allocate_action_node (struct list_s *node, char *bptr)
{
  struct action *payload;
  int i;

  /* Allocate action node */
  node->payload = (void *) malloc (sizeof (struct action));
  if (node->payload == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  payload = (struct action *) node->payload;
  payload->name = (char *) calloc (strlen (bptr) + 1, sizeof (char));
  if (payload->name == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  /* Init action members */
  payload->ignore = d_array_create ();
  payload->run_action = NULL;
  payload->id = EMPTY;
  payload->ltvalue = ALWAYS;

  strncpy (payload->name, bptr, strlen (bptr));

  /* By default all substrings are used to indentify event 
   * (fill lookup table with 1's) */
  for (i = 0; i < SUBSTRMAX; i++)
    payload->usesubstr[i] = 1;

  return payload;
}

static void
initialize_action (struct action *acell)
{
  int index;
  
  /* We need to create array for events */
  if (acell->run_action == NULL)
    acell->run_action = d_array_create ();

  /* Add new element in run_action array */
  index = d_array_add_element (acell->run_action);

  /* Allocate space for payload */
  acell->run_action->element[index]->payload = (void *)
    malloc (sizeof (struct do_action));
  if (acell->run_action->element[index]->payload == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  /* Initialize run_action members */
  ((struct do_action *) acell->run_action->element[index]->payload)->action = NULL;
  ((struct do_action *) acell->run_action->element[index]->payload)->t_limit = -1;
  ((struct do_action *) acell->run_action->element[index]->payload)->event_max = -1;

  action_allocated = 1;
}

/* Arrange rule list nodes aptr pointer to correct action nodes */
static void
arrange_nodes (struct list_s **rlist, struct list_s **alist)
{
  struct list_s *rcur;
  struct list_s *acur;
  struct rule *rdata;
  struct action *adata;

  rcur = *rlist;

  while (rcur != NULL)
    {
      rdata = (struct rule *) rcur->payload;
      acur = *alist;

      while (acur != NULL)
	{
	  adata = (struct action *) acur->payload;
	  if (!strncmp (rdata->actname, adata->name, strlen (rdata->actname)))
	    {
	      rdata->aptr = (struct action *) acur->payload;
              free (rdata->actname);
              rdata->actname = NULL;
	      break;
            }
          else
            rdata->aptr = NULL;
	  acur = acur->next;
	}

      if (rdata->aptr == NULL)
	{
	  fprintf (stderr,
		   "Error: can't find %s action function for rule #%d\n",
		   rdata->actname, rdata->rulenr);
	  exit (1);
	}
      rcur = rcur->next;
    }
}

/* This function parse multiple lines for ignore list
 * and allocates/initializes structures for it
 */
static int
parse_ignorelist (struct d_array *ignore, char *buffer)
{
  char *cptr;
  char *element;
  char pname[4096] = { 0 };
  int index;

  cptr = &buffer[0];
  element = &pname[0];

  while (isgraph ((int) *cptr) || *cptr == '/')
    *element++ = *cptr++;
  while (isspace ((int) *cptr))
    *cptr++;

  index = d_array_add_element (ignore);

  ignore->element[index]->payload = (void *) calloc (strlen (pname) + 1, sizeof (char));
  if (ignore->element[index]->payload == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  strcpy ((char *) ignore->element[index]->payload, pname);

  if (*cptr == '\\')
    return 0;
  else
    return 1;
}

/* Build lookup table for substrings. Table marks substring
 * indexes which are used to indentify event
 */
static int
do_substr_table (struct action *new_node, char *bptr)
{
  int i;
  int idx;

  for (i = 0; i < SUBSTRMAX; i++)
    new_node->usesubstr[i] = 0;

  for (i = 0; i < strlen (bptr); i++)
    {
      if (bptr[i] != '$' && !isspace ((int) bptr[i]) && !isdigit ((int) bptr[i]))
	return -1;
      if (bptr[i - 1] == '$' && isdigit ((int) bptr[i]))
	{
	  idx = atoi (&bptr[i]);
	  new_node->usesubstr[idx] = 1;
	}
    }
  return 0;
}
